import math
import numpy as np
import cartcoords
from math import pi

class ZItem:
	def __init__(self, value=None, dist=float('-inf'), index1=-1, angle=float('-inf'), index2=-1, torsion=float('-inf'), index3=-1):
		#assert( (dist!=float('-inf')) == (index1>=0) )
		#assert( (angle!=float('-inf')) == (index2>=0) )
		#assert( (torsion!=float('-inf')) == (index3>=0) )
		self.value	=value
		self.dist	=dist		# > 0
		self.index1	=index1
		self.angle	=angle		# 0..180
		self.index2	=index2
		self.torsion	=torsion	# 0 .. 360
		self.index3	=index3
	
	def round(self, prec):
		return ZItem(self.value, round(self.dist,prec), self.index1, round(self.angle,prec), self.index2, round(self.torsion, prec), self.index3)
		
	def geom_only(self):
		return ZItem(dist=self.dist, angle=self.angle, torsion=self.torsion)
		
	def __cmp__(self, other):
		return cmp((self.value, self.dist, self.index1, self.angle, self.index2, self.torsion, self.index3), (other.value, other.dist, other.index1, other.angle, other.index2, other.torsion, other.index3))
		
	def __repr__(self):
		return "{val}  {dist} {i1} {angle} {i2} {torsion} {i3}".format(val=self.value, dist=self.dist, i1=self.index1, angle=180.0*self.angle/math.pi, i2=self.index2, torsion=180.0*self.torsion/math.pi, i3=self.index3)
		
	def __hash__(self):
		return hash((self.value,self.dist,self.index1,self.angle,self.index2,self.torsion,self.index3))

	def geom_cmp(self, zi_right, precission):
		self_round = self.round(precission).geom_only()
		zi_right_round = zi_right.round(precission).geom_only()
	
		return cmp(self_round, zi_right_round)

def cartesian2zItem(u,r1,r2,r3):
	v1 = (u-r1)
	dist = np.linalg.norm(v1)	
	if r2 == None:
		return ZItem(dist=dist)
		
	v2 = (r2-r1)
	
	cos_a = np.dot(v1 * 1.0/dist, v2 * 1.0/np.linalg.norm(v2))
	assert(abs(cos_a) < 1.0000001)
	#print str(v1) + " " + str(v2)
	#print str(v1 * 1.0/dist) + " " + str(v2 * 1.0/np.linalg.norm(v2))
	if (cos_a > 1):
		cos_a = 1
	if (cos_a < -1):
		cos_a = -1	

	angle = math.acos(cos_a)
	if r3 == None:
		return ZItem(dist=dist, angle=angle)
	

	b1 = (r1-u)
	b2 = (r2-r1)
	b3 = (r3-r2)
	
	n1 = np.cross(b1,b2)
	n2 = np.cross(b2,b3)
	
	torsion = math.atan2(np.dot(np.cross(n1,n2), b2)/np.linalg.norm(b2), np.dot(n1,n2))
	if torsion < 0:
		torsion += math.pi*2
	if torsion > math.pi*1.99:
		torsion = 0.0
	
	return ZItem(dist=dist, angle=angle, torsion=torsion)
	
def make_single_point_zmatrix():
	zi0 = ZItem(value='1') #red
	return [zi0]

def make_dimer_zmatrix(a):
	zi0 = ZItem(value='1') #red
	zi1 = ZItem(value='2', dist=a, index1=0) #yellow
	return [zi0, zi1]

def make_trimer_zmatrix(a1,a2,angle):
	z01 = make_dimer_zmatrix(a1)
	zi2 = ZItem(value='3', dist=a2, index1=0, angle=angle, index2=1) 
	return z01 + [zi2]

def make_tetrahedron_zmatrix(a):
	z012 = make_trimer_zmatrix(a,a,pi/3)
	zi3 = ZItem(value='4', dist=a, index1=0, angle=pi/3, index2=1, torsion = math.acos(1./3.), index3=2) 
	return z012 + [zi3]

def make_trimer_zmatrix_2(a1,a2,angle):
	z01 = make_dimer_zmatrix(a1)
	zi2 = ZItem(value='3', dist=a2, index1=0, angle=angle, index2=1)
	
	
	a = a1/(3.0**0.5)
	
	zi3 = ZItem(value='4', dist=a, index1=1, angle=(pi-angle)/2, index2=0, torsion=0.0, index3=2)
	zi4 = ZItem(value='5', dist=a, index1=2, angle=(pi-angle)/2, index2=0, torsion=0.0, index3=1)
	return z01 + [zi2,zi3,zi4]


def make_trapez(a,angle):
	z02 = make_trimer_zmatrix(a,a,angle)

	zi3 = ZItem(value='4', dist=a, index1=1, angle=angle, index2=0, torsion=pi, index3=2)
	zi4 = ZItem(value='5', dist=a, index1=2, angle=angle, index2=0, torsion=pi, index3=1)
	

	return z02 + [zi3,zi4]

"""	
def make_fcc_base_zmatrix(a):
	zi0 = ZItem(value='1') #red
	zi1 = ZItem(value='2', dist=3., index1=0) #yellow
	zi2 = ZItem(value='3', dist=3., index1=1, angle=60./180.*math.pi, index2=0) #purple
	return [zi0, zi1, zi2]
	
def make_cubic_base_zmatrix(a):
	zi0 = ZItem(value='1') #red
	zi1 = ZItem(value='2', dist=3., index1=0) #yellow
	zi2 = ZItem(value='3', dist=3., index1=1, angle=90./180.*math.pi, index2=0) #purple
	return [zi0, zi1, zi2]	
"""
	
def test_cartesian2zItem():
	'''
	Pt         0.00000        0.00000        0.00000
	Pt         2.74490        0.00000        0.00000
	Pt         1.37245        0.00000       -2.37715
	Pt         1.37245        2.24122       -0.79234
	'''
	zitem_orig = ZItem(dist=2.7449, angle=60.0/180.0*math.pi, torsion=289.47/180.0*math.pi).round(3)
	zitem_calc = cartesian2zItem(np.array([1.37245,2.24122,-0.79234]), np.array([0,0,0]), np.array([2.7449,0,0]), np.array([1.37245,0,-2.37715])).round(3)
	#print str(zitem_orig) + "=?=" + str(zitem_calc)
	assert(zitem_orig == zitem_calc)
	


